/*
 * channels constructor / destructor API
 *
 * Copyright (c) 2012, LiteStack, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef CHANNEL_H_
#define CHANNEL_H_

#include <sys/un.h>
#include "api/zvm.h"
#include "src/main/etag.h"
#include "src/main/manifest.h"

EXTERN_C_BEGIN

/* net buffer size depends on it */
#define BUFFER_SIZE 0x10000
#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX 108
#endif

/* zerovm channels names */
#define STDIN "/dev/stdin"
#define STDOUT "/dev/stdout"
#define STDERR "/dev/stderr"
#define STDRAM "/dev/memory"

#define FLAG_VALID_MASK 8
#define IS_NETWORK(c) ((c)->protocol < ProtoRegular)
#define IS_FILE(c) (!IS_NETWORK(c))
#define IS_IPHOST(c) ((c)->flags & 1)
#define IS_VALID(c) (!((c)->flags & FLAG_VALID_MASK))

/* CH_RW_TYPE returns 0..3 */
#define IS_NIL(channel) (CH_RW_TYPE(channel) == 0)
#define IS_RO(channel) (CH_RW_TYPE(channel) == 1)
#define IS_WO(channel) (CH_RW_TYPE(channel) == 2)
#define IS_RW(channel) (CH_RW_TYPE(channel) == 3)
#define CH_RW_TYPE(channel) \
    (((channel)->limits[GetsLimit] && (channel)->limits[GetSizeLimit]) \
    | ((channel)->limits[PutsLimit] && (channel)->limits[PutSizeLimit]) << 1)

#define CH_CH(manifest, n) (((struct ChannelDesc**)(manifest->channels->pdata))[n])
#define CH_FILE(channel, n) ((struct File*)channel->source->pdata[n])
#define CH_CONN(channel, n) ((struct Connection*)channel->source->pdata[n])
#define CH_HANDLE(channel, n) (CH_FILE(channel, n))->handle
#define CH_PROTO(channel, n) (CH_FILE(channel, n))->protocol
#define CH_FLAGS(channel, n) (CH_FILE(channel, n))->flags
#define CH_NAME(channel, n) (CH_FILE(channel, n))->name
#define CH_HOST(channel, n) (CH_CONN(channel, n))->host
#define CH_PORT(channel, n) (CH_CONN(channel, n))->port
#define CH_BACKUP(channel, n) (CH_CONN(channel, n))->backup

#define CH_SEQ_READABLE(channel) (((channel)->type & 1) == 0)
#define CH_SEQ_WRITEABLE(channel) (((channel)->type & 2) == 0)
#define CH_RND_READABLE(channel) (((channel)->type & 1) == 1)
#define CH_RND_WRITEABLE(channel) (((channel)->type & 2) == 2)



/*@

  predicate valid_limit{L}(struct ChannelDesc *ch, integer i) =
       ch->counters[i] <= ch->limits[i]
       && 0 <= ch->counters[i]
       && 0 <= ch->limits[i];

  predicate valid_channel{L}(struct ChannelDesc *ch) =
  \valid(ch) && valid_string(ch->alias) && ch->size >= 0
  && ch->getpos >= 0
  && \valid(ch->limits+(0..LimitsNumber))
  && \valid(ch->counters+(0..LimitsNumber))
  && (\forall integer i; (0 <= i && i < LimitsNumber) ==> valid_limit(ch,i));

  predicate valid_nap_aux{L}(struct NaClApp *nap) =
  \valid(nap) && \valid(nap->manifest) &&
  0 < nap->addr_bits &&
  nap->addr_bits <= NACL_MAX_ADDR_BITS &&
  \valid(nap->manifest->channels) &&
  \valid(((struct ChannelDesc**)nap->manifest->channels->pdata)
  +(0..nap->manifest->channels->len-1)) &&

  (\forall integer i; 0 <= i && i < nap->manifest->channels->len ==>
    valid_channel(CH_CH(nap->manifest,i))) &&

  \valid(nap->mem_map+(0..MemMapSize)) &&
  (\forall integer i; 0 <= i && i < MemMapSize ==>
  valid_memblock(nap->mem_map[i])) &&
  (\forall integer i; 0 <= i && i < MemMapSize - 1 ==>
      nap->mem_map[i].end == nap->mem_map[i+1].start) &&
  nap->mem_start + INT32_MAX <= INT64_MAX &&
  nap->mem_map[LeftBumperIdx].start < nap->mem_start &&
  nap->mem_map[LeftBumperIdx].prot == PROT_NONE &&
  nap->mem_map[RightBumperIdx].prot == PROT_NONE &&
  nap->data_end <= nap->break_addr;

  axiomatic color {
    logic integer color_pointer(void * p);
  }


  predicate nap_sep{L}(struct NaClApp *nap) =
     color_pointer(\base_addr(nap)) == -2 &&
     color_pointer(\base_addr(nap->manifest->channels)) == -1 &&
     color_pointer(\base_addr(nap->manifest)) == -3 &&
     color_pointer(\base_addr(nap->manifest->channels->pdata)) == -4 &&

     (\forall integer i; 0 <= i && i < nap->manifest->channels->len ==>
     color_pointer(\base_addr(CH_CH(nap->manifest,i))) == i);


  lemma nap_sep_frame{L1,L2}:
  \forall struct NaClApp *nap; \forall integer i;
  \let ch = \at(CH_CH(nap->manifest,i),L1);
  nap_sep{L1}(nap) ==>
  valid_nap_aux{L1}(nap) ==>
  \at(0 <= i && i < nap->manifest->channels->len,L1) ==>

  (\forall char *c; \at(\valid(c),L1) <==> \at(\valid(c),L2)) ==>
  (\forall gpointer *c; \at(\valid(c),L1) <==> \at(\valid(c),L2)) ==>
  (\forall int  *c; \at(\valid(c),L1) <==> \at(\valid(c),L2)) ==>

  (\forall char *c; \at(\separated(c,ch),L1) ==> \at(*c,L1) == \at(*c,L2)) ==>
  (\forall gpointer *c; \at(\separated(c,ch),L1) ==> \at(*c,L1) == \at(*c,L2)) ==>
  (\forall int  *c; \at(\separated(c,ch),L1) ==> \at(*c,L1) == \at(*c,L2)) ==>

  \at(validity_memory_state,L1) == \at(validity_memory_state,L2) ==>
  valid_channel{L2}(ch) ==>
  (valid_nap_aux{L2}(nap) && nap_sep{L2}(nap));

  predicate valid_nap{L}(struct NaClApp *nap) =
    valid_nap_aux(nap) && nap_sep(nap);

@*/

/* sort channels */
void SortChannels(GPtrArray *channels);

/* construct all channels, initialize it and update system_manifest */
void ChannelsCtor(struct Manifest *manifest);

/* free channels resources */
void ChannelsDtor(struct Manifest *manifest);

/*@
axiomatic results{
 predicate read_result{L}(struct ChannelDesc *channel,
 size_t size, off_t offset, int32_t result)
 reads channel;

 predicate write_result{L}(struct ChannelDesc *channel,
 size_t size, off_t offset, int32_t result)
 reads channel;
}
 @*/

/* read channel data through multiple sources */
/*@
  requires valid_channel(channel);
  requires valid_segment(buffer,buffer+size);
  requires offset >= 0;
  ensures valid_channel(channel);
  ensures read_result(channel, size, offset, \result);
  assigns channel->counters[..];
  @*/
int32_t ChannelRead(struct ChannelDesc *channel,
    NaClSysPtr buffer, size_t size, off_t offset);

/* write channel data through multiple sources */
/*@
  requires valid_channel(channel);
  requires valid_read_segment(buffer,buffer+size);
  requires offset >= 0;
  ensures valid_channel(channel);
  ensures write_result(channel, size, offset, \result);
  assigns channel->counters[..];
  @*/
int32_t ChannelWrite(struct ChannelDesc *channel,
    NaClSysPtr buffer, size_t size, off_t offset);

EXTERN_C_END

#endif /* CHANNEL_H_ */
